/* -----------------------------------------------------------------
Licence Agreement

You have a royalty-free right to use, modify, reproduce and
distribute this program in any way you find useful, provided that
you agree that Bill Buckels has no warranty obligations or
liability resulting from said distribution in any way whatsoever.
If you don't agree, remove this source code and its associated
files from your computer now.

System       : MS-DOS (etc)
Program      : metasim.c
Written by   : Bill Buckels
               589 Oxford Street
               Winnipeg, Manitoba, Canada R3M 3J2

Email: bbuckels@escape.ca
WebSite: http://www.escape.ca/~bbuckels

Description  : A Filter-program to host Similarity Sort Key Creation
               With or Without Metaphone Algorithm

               A Similarity Sort Key is Created With Metaphone
               Switch Setting None (this is the default)
               or *optionally*
               A Similarity Sort Key is Created Without Metaphone,
               Switch Setting -v (Verbatim)

               and *optionally*
               An Unordered Similarity Sort Key Is Created
               Switch Setting None (this is the default)
               or *optionally*
               An Ordered Similarity Sort Key is Created
               for reverse equivalent for phrase comparison
               Switch Setting -r (Reverse)

               and *optionally*
               The Sort Key Precedes The Output so an
               External Sort Can Be Used...
               Switch Setting None (this is the default)
               or *optionally*
               Sorts Short Lists (approx. 2900 entries),
               using an Internal Sorting Algorithm
               without the need for an external sort,
               and without the need for external cleanup (lchop).
               Switch Setting -s (Sort)

               Input from stdout, stream, file, or command line, or
               redirect from file.

               Output to stdout, stream, file, or pipe through other
               filters, or redirect to file.

Based On     : Metaphone Algorithm

               Created by Lawrence Philips.
               Metaphone presented in article in "Computer Language"
               December 1990.

               Converted from Pick BASIC, as demonstrated in article,
               to C by Michael J. Kuhn (Baltimore, Maryland)

               Massage for Similiarity Sort added by Bill Buckels.

Date Written : July 2000
Revision     : 1.1 First Release
                   Basic Version To the Fidonet C_ECHO
                   - Metaphone Sort Key Only
                   - Reverse Equivalent Sort Key Only
                   - Internal Dictionary HardCoded
               1.2 Second Release August 2000
                   Added Options
                   - Internal Sort or External Sort
                   - Reverse Equivalent or Natural Order Key
                   - Verbatim or Metaphone Sort
                   Externalized Dictionary
                   - Allow Synonyms
                   - Allow Word Removal From Sort Key
                   - Allow Prefixes
                   - Allow Suffixes
                   - Permits Other Languages Besides English

Written in Large Model Microsoft C Version 6.00a

----------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* general purpose constants */

/* input modes */
enum {
  INPUT_STREAM,
  INPUT_FILE,
  INPUT_ARGV};

/* dictionary sections */
#define NO_SECTION (-1)

/* section id's */
enum {
  PREFIX_SECTION = (NO_SECTION+1),
  SUFFIX_SECTION,
  SUB_SECTION,
  REMOVE_SECTION,
  NUM_SECTIONS};

int CurrentSection = NO_SECTION;

/* section names */
char *Section[] = {
  "[prefix]",
  "[suffix]",
  "[sub]",
  "[remove]"};

/* character constants */
#define CARRIAGE_RETURN '\r'
#define LINE_FEED       '\n'
#define ASCIIZ          '\x00'
#define CONTROL_Z       '\x1a'
#define SPACE           ' '

/* buffer size constants */
#define NAME_BUF_LEN    1024
#define META_BUF_LEN    50
#define META_LEN        20
#define LEN_LEN         sizeof(int)
#define POS_LEN         sizeof(long)
#define WORD_LEN        128
#define WORD_MAX        100
#define DIC_LEN         40
#define KEY_LEN         (META_LEN+POS_LEN)
#define KEY_MAX         (((unsigned)(64000))/KEY_LEN)

/* global buffers - to reduce stress on stack */

char szName[NAME_BUF_LEN];
char szMetaph[META_BUF_LEN];
char szSimBuf[NAME_BUF_LEN];

char *pszCredit =
  "MetaSim(C) v1.2 CopyLeft Bill Buckels 2000. All Rights Reversed.\n"
  "\tMetaphone Algorithm Created by Lawrence Philips.\n"
  "\tConverted to C by Michael J. Kuhn.\n"
  "\tFilter to host Metaphone Algorithm and\n"
  "\tSimilarity Sort Key Creation Algorithm by Bill Buckels.\n";

char *pszPurpose =
  "Purpose:  Metaphone Filter with Similarity Sort.\n"
  "\tInput from stdin, stream, file, or command line.\n"
  "\tOutput to stdout, stream, file.\n";

char *pszUsage   =
  "Usage(s): metasim\n"
  "\tmetasim [infile.lst]\n"
  "\tmetasim [infile.lst] [outfile.lst]\n"
  "\tmetasim [phrase1] [phrase2] [phrase3] ...\n"
  "\tmetasim < [infile.lst]\n"
  "\ttype [infile.lst] | metasim > [outfile.lst]\n"
  "Switches: metasim -rsvd [infile.lst] ...\n"
  "\tr - reverse equivalent comparison rather than natural.\n"
  "\ts - sort internally and trim key from output.\n"
  "\tv - verbose sort key rather than Metaphone key.\n"
  "\td - debug output (use internal sort but don't strip keys).\n";

char *pszPrompt =
  "Enter string(s) to be converted (Enter CTRL-Z to quit):\n";

/* external files used by metasim */
char szTempSortFile[128] = "METASIM.$$$";
char szDicFile[128]      = "METASIM.DIC";

/* metaphone algorithm specific constants and vars */

#define TRUE  (1)
#define FALSE (0)
#define NULLCHAR (char *)0

char *VOWELS="AEIOU",
     /* special cases for letters in FRONT of these */
     *FRONTV="EIY",
     /* variable sound--those modified by adding an "h" */
     *VARSON="CSPTG",
     /* let these double letters through */
     *DOUBLE=".";

     /* exceptions "ae-", "gn-", "kn-", "pn-", "wr-" */
char *excpPAIR="AGKPW",
     *nextLTR ="ENNNR";
char *chrptr, *chrptr1;

/*** start of metaphone algorithm ***/

/* -------------------------------------------------------------- */
/* function phonetic                                              */
/* implementation of metaphone - by Michael J. Kuhn               */
/* input(s) : inputbuffer, outputbuffer, outputlength             */
/* output(s): outputbuffer (metaphone equivalent of inputbuffer)  */
/* returns  : none                                                */
/* -------------------------------------------------------------- */
void phonetic(char *name, char *metaph, int metalen)
{
  int  ii, jj, silent, hard, Lng, lastChr,
       vowelAfter, vowelBefore, frontvAfter;
  char curLtr, prevLtr, nextLtr, nextLtr2, nextLtr3,
       wname[NAME_BUF_LEN],
       *ename = wname;
  
  metaph[0] = ASCIIZ; /* terminate output buffer */

  /* work on a copy - alpha only */
  jj = 0;
  for(ii = 0; name[ii] != ASCIIZ; ii++)
  {
    if(isalpha(name[ii]))
    {
      ename[jj] = toupper(name[ii]);
      jj++;
    }
  }
  ename[jj] = ASCIIZ;
  if(strlen(ename) == 0) return; /* if no alpha just return */

  /* if ae, gn, kn, pn, wr then drop the first letter */
  if((chrptr = strchr(excpPAIR, ename[0]) ) != NULLCHAR) 
  {
    chrptr1 = nextLTR + (chrptr - excpPAIR);
    if(*chrptr1 == ename[1])  strcpy(ename, &ename[1]);
  }
  
  /* change x to s */
  if(ename[0] == 'X') ename[0] = 'S';
  
  /* get rid of the "h" in "wh" */
  if(strncmp(ename, "WH", 2) == 0) strcpy(&ename[1], &ename[2]);
  
  Lng = strlen(ename);
  /* index to last character in string makes code easier*/
  lastChr = Lng - 1;
  
  /* Remove an S from the end of the string */
  if(ename[lastChr] == 'S') 
  {
    ename[lastChr] = ASCIIZ;
    Lng = strlen(ename);
    lastChr = Lng - 1;
  }

  for(ii = 0; ( (strlen(metaph) < metalen) && (ii < Lng) ); ii++) 
  {
    curLtr = ename[ii];
    
    vowelBefore = FALSE;  prevLtr = SPACE;
    if(ii > 0) 
    {
      prevLtr = ename[ii - 1];
      if(strchr(VOWELS, prevLtr) != NULLCHAR) vowelBefore = TRUE;
    }
    /* if first letter is a vowel KEEP it */
    if(ii == 0 && (strchr(VOWELS, curLtr) != NULLCHAR)) 
    {
      strncat(metaph, &curLtr, 1);
      continue;
    }

    vowelAfter = FALSE;   frontvAfter = FALSE;   nextLtr = SPACE;
    if(ii < lastChr) 
    {
      nextLtr = ename[ii + 1];
      if(strchr(VOWELS, nextLtr) != NULLCHAR) vowelAfter = TRUE;
      if(strchr(FRONTV, nextLtr) != NULLCHAR) frontvAfter = TRUE;
    }
    /* skip double letters except ones in list */
    if(curLtr == nextLtr &&
       (strchr(DOUBLE, nextLtr) == NULLCHAR)) continue;

    nextLtr2 = SPACE;
    if(ii < (lastChr - 1)) nextLtr2 = ename[ii + 2];
    
    nextLtr3 = SPACE;
    if(ii < (lastChr - 2)) nextLtr3 = ename[ii + 3];
    
    switch(curLtr) 
    {
      case 'B':
        silent = FALSE;
        if(ii == lastChr && prevLtr == 'M') silent = TRUE;
        if(! silent) strncat(metaph, &curLtr, 1);
        break;

      /*silent -sci-,-sce-,-scy-;  sci-, etc OK*/
      case 'C':
        if(! (ii > 1 && prevLtr == 'S' && frontvAfter)) {
          if(ii > 0 && nextLtr == 'I' && nextLtr2 == 'A') {
            strncat(metaph, "X", 1);
          }
          else {
            if(frontvAfter) {
              strncat(metaph, "S", 1);
            }
            else {
              if(ii > 1 && prevLtr == 'S' && nextLtr == 'H') {
                strncat(metaph, "K", 1);
              }
              else {
                if(nextLtr == 'H') {
                  if(ii == 0 &&
                     (strchr(VOWELS, nextLtr2) == NULLCHAR))
                    strncat(metaph, "K", 1);
                  else
                    strncat(metaph, "X", 1);
                }
                else  {
                  if(prevLtr == 'C') {
                    strncat(metaph, "C", 1);
                  }
                  else {
                    strncat(metaph, "K", 1);
                  }
                }
              }
            }
          }
        }
        break;
      
      case 'D':
        if(nextLtr == 'G' &&
           (strchr(FRONTV, nextLtr2) != NULLCHAR))
          strncat(metaph, "J", 1);
        else
          strncat(metaph, "T", 1);
        break;
      
      case 'G': silent = FALSE;
        /* SILENT -gh- except for -gh and no vowel after h */
        if((ii < (lastChr - 1) && nextLtr == 'H')
           && (strchr(VOWELS, nextLtr2) == NULLCHAR))
          silent = TRUE;
      
        if((ii == (lastChr - 3) ) &&
           nextLtr == 'N' && nextLtr2 == 'E' && nextLtr3 == 'D')
          silent = TRUE;
        else
          if((ii == (lastChr - 1)) && nextLtr == 'N')
            silent = TRUE;
      
        if(prevLtr == 'D' && frontvAfter) silent = TRUE;

        if(prevLtr == 'G')
          hard = TRUE;
        else
          hard = FALSE;
      
        if(!silent) {
          if(frontvAfter && (! hard))
            strncat(metaph, "J", 1);
          else
            strncat(metaph, "K", 1);
        }
        break;
      
      case 'H':
        silent = FALSE;
        if(strchr(VARSON, prevLtr) != NULLCHAR) silent = TRUE;
        if(vowelBefore && !vowelAfter) silent = TRUE;
        if(!silent) strncat(metaph, &curLtr, 1);
        break;
      
      case 'F':
      case 'J':
      case 'L':
      case 'M':
      case 'N':
      case 'R':
        strncat(metaph, &curLtr, 1);
        break;
      
      case 'K':
        if(prevLtr != 'C') strncat(metaph, &curLtr, 1);
        break;
      
      case 'P':
        if(nextLtr == 'H')
          strncat(metaph, "F", 1);
        else
          strncat(metaph, "P", 1);
        break;
      
      case 'Q':
        strncat(metaph, "K", 1);
        break;
      
      case 'S':
        if(ii > 1 && nextLtr == 'I' &&
           ( nextLtr2 == 'O' || nextLtr2 == 'A'))
          strncat(metaph, "X", 1);
        else
          if(nextLtr == 'H')
            strncat(metaph, "X", 1);
          else
            strncat(metaph, "S", 1);
        break;
      
      case 'T':
        if(ii > 1 && nextLtr == 'I' &&
           (nextLtr2 == 'O' || nextLtr2 == 'A')) {
          strncat(metaph, "X", 1);
        }
        else {
          if(nextLtr == 'H') { /* The=0, Tho=T, Withrow=0 */
            if(ii > 0 || (strchr(VOWELS, nextLtr2) != NULLCHAR))
              strncat(metaph, "0", 1);
            else
              strncat(metaph, "T", 1);
          }
          else  {
            if(! (ii < (lastChr - 2) &&
               nextLtr == 'C' && nextLtr2 == 'H'))
              strncat(metaph, "T", 1);
          }
        }
        break;
      
      case 'V':
        strncat(metaph, "F", 1);
        break;
      
      case 'W':
      case 'Y':
        if(ii < lastChr && vowelAfter)
          strncat(metaph, &curLtr, 1);
        break;
      
      case 'X':
        strncat(metaph, "KS", 2);
        break;
      
      case 'Z':
        strncat(metaph, "S", 1);
        break;
    }
    
  }
  return;
}

/*** end of metaphone algorithm ***/

/* start of word re-sort (set-up area for Similarity Sort Key */
/* and other optional features of this program */

/* 1. flag for reverse equivalent for phrase comparison */
/* set to false unless a switch is used */
/* 2. flag for internal sort set to false unless a switch is used */
/* 3. flag to turn-off metaphone algorithm and use a "squished key" */
/* instead... by default, set to false unless a switch is used */
/* by default metaphone is on */

int bReverse = FALSE,
    bSort = FALSE,
    bVerbatim = FALSE,
    bDebug = FALSE;

/* structure for word re-sort */
typedef struct {
  char szSimWord[WORD_LEN+1];
} WILDWORD;

/* structure for suffix, prefix, remove */
typedef struct {
  char szSimDic[DIC_LEN+1];
} WILDDIC;

/* structure for synonym */
typedef struct {
  char szFind[DIC_LEN+1];
  char szReplace[DIC_LEN+1];
} WILDSUB;

/* structure for internal sort */
typedef struct {
  char szSortKey[META_LEN];  /* key value */
  long lpos;                /* offset in temporary file */
} WILDPHRASE;

WILDWORD *wildword = (WILDWORD *)NULL;

WILDPHRASE *wildphrase = (WILDPHRASE *)NULL;

/* compare function for quicksort */

unsigned int sortsize = sizeof(WILDWORD);
int compare(void **arg1, void **arg2)
{
  return(memicmp(arg1, arg2, sortsize));
}

/* prefixes, suffixes, and words */
/* that are ignored when building the */
/* similarity sort key */
/* user definable and externalized... */

/* these are stored in the dictionary file */
/* Metasim.dic which must be in the current directory */
/* add or remove as necessary... */

/* please see the comments in the sample Metasim.dic */
/* provided with Metasim for additional details. */

unsigned int numsubs = 0,
    numskips = 0,
    numprefixes = 0,
    numsuffixes = 0;

WILDSUB  *sub     = (WILDSUB *)NULL;
WILDDIC  *skip    = (WILDDIC *)NULL;
WILDDIC  *prefix  = (WILDDIC *)NULL;
WILDDIC  *suffix  = (WILDDIC *)NULL;

/* -------------------------------------------------------------- */
/* function SubAndRemove                                          */
/* Helper Function called by MassageSimilaritySortWord            */
/* substitutes words that need to be substituted                  */
/* removes words that need to be ignored                          */
/* called before prefix removal, before suffix removal, and       */
/* at the end of prefix-suffix removal                            */
/* input(s) : outputbuffer                                        */
/* output(s): outputbuffer (modified)                             */
/* returns  : returns buffer string length                        */
/* -------------------------------------------------------------- */
int SubAndRemove(char *szOutBuf)
{
  int idx;

  /* search and replace first */
  if (szOutBuf[0] != ASCIIZ) {
    for (idx = 0; idx < numsubs; idx++) {
      if (!stricmp(szOutBuf, (char *)&sub[idx].szFind[0])) {
        /* replace the old word with the new word */
        strcpy(szOutBuf, (char *)&sub[idx].szReplace[0]);
        break;
      }
    }
  }

  /* skip words second */
  if (szOutBuf[0] != ASCIIZ) {
    for (idx = 0; idx < numskips; idx++) {
      if (!stricmp(szOutBuf, (char *)&skip[idx].szSimDic[0])) {
        /* truncate buffer to remove the word */
        szOutBuf[0] = ASCIIZ;
        break;
      }
    }
  }

  if (szOutBuf[0] == ASCIIZ) return (0);
  return (strlen(szOutBuf));

}


/* -------------------------------------------------------------- */
/* function MassageSimilaritySortWord                             */
/* Helper Function called by CreateSimilaritySortKey              */
/* checks for words that need to be ignored and                   */
/* removes suffixes that need to be ignored                       */
/* input(s) : outputbuffer                                        */
/* output(s): outputbuffer (modified)                             */
/* returns  : 0 if false otherwise returns buffer string length   */
/* -------------------------------------------------------------- */
int MassageSimilaritySortWord(char *szOutBuf)
{
  int idx, jdx, len, pos;

  /* removal pass 1 */
  SubAndRemove(szOutBuf);

  /* check prefixes */
  if (szOutBuf[0] != ASCIIZ) {
    len = strlen(szOutBuf);
    for (idx = 0; idx < numprefixes; idx++) {
      pos = strlen((char *)&prefix[idx].szSimDic[0]);
      if (pos > len)
        continue;
      if (!memcmp(szOutBuf, (char *)&prefix[idx].szSimDic[0], pos)) {
        /* ltrim the prefix from the start of the string */
        for (jdx =0;;jdx++, pos++) {
          szOutBuf[jdx] = szOutBuf[pos];
          if (szOutBuf[pos] == ASCIIZ)
            break;
        }
        break;
      }
    }
  }

  /* removal pass 2 */
  SubAndRemove(szOutBuf);

  /* check suffixes */
  if (szOutBuf[0] != ASCIIZ) {
    len = strlen(szOutBuf);
    for (idx = 0; idx < numsuffixes; idx++) {
      pos = len - strlen((char *)&suffix[idx].szSimDic[0]);
      if (pos < 0)
        continue;
      if (!stricmp(&szOutBuf[pos], (char *)&suffix[idx].szSimDic[0])) {
        /* rtrim the suffix from the end of the string */
        szOutBuf[pos] = ASCIIZ;
        break;
      }
    }
  }

  /* removal pass 3 */
  SubAndRemove(szOutBuf);

  if (szOutBuf[0] == ASCIIZ) return (0);
  return (strlen(szOutBuf));

}


/* -------------------------------------------------------------- */
/* function CreateSimilaritySortKey                               */
/* Reformats The Input String into a sorted string with certain   */
/* words removed (see function MassageSImiliaritySortWord)        */
/* Returns a reformatted Copy which is used to call MetaPhone     */
/* instead of the "real" string...                                */
/* input(s) : inputbuffer, outputbuffer                           */
/* output(s): outputbuffer (modified)                             */
/* returns  : returns number of words in output buffer            */
/* -------------------------------------------------------------- */
int CreateSimilaritySortKey(char *szInBuf, char *szOutBuf)
{
    int cnt = 0,         /* outstring   */
        idx = 0,         /* instring    */
        elements = 0;    /* total words */
    char c;

    /* break the phrase into individual words */
    for (;;) {
      c = toupper(szInBuf[idx]);
      if (c==ASCIIZ || c==CARRIAGE_RETURN || c==LINE_FEED)
        break;
      idx++;

      /* alpha only except for dashes and apostrophes */
      if (isalpha(c) || c=='-'|| c=='\'') {
        szOutBuf[cnt] = c;
        cnt++;
        szOutBuf[cnt] = ASCIIZ;
        continue;
      }
      /* non alpha is the end of the word */
      if (cnt > 0) {
        if (MassageSimilaritySortWord(szOutBuf)) {
          strcpy(wildword[elements].szSimWord, szOutBuf);
          elements++;
        }
        cnt = 0;
      }
    }

    /* stragglers */
    if (cnt > 0) {
      if (MassageSimilaritySortWord(szOutBuf)) {
         strcpy(wildword[elements].szSimWord, szOutBuf);
         elements++;
      }
    }

    /* if we are using the reverse equivalent sort */
    /* then qsort if we have any words to sort */

    if (elements > 1 && bReverse == TRUE)
      qsort(wildword, elements, sizeof(WILDWORD), compare);

    /* gather if we have any words to gather */
    szOutBuf[0] = ASCIIZ;
    for (idx = 0; idx < elements; idx++)
      strcat(szOutBuf, wildword[idx].szSimWord);

    return (elements);
}

/* parsing area for reading of the dictionary */

/* -------------------------------------------------------------- */
/* function StringUpper                                           */
/* Creates Upper Case Copy of A String                            */
/* input(s) : src, dest                                           */
/* output(s): dest (upper)                                        */
/* returns  : none                                                */
/* -------------------------------------------------------------- */
void StringUpper(char *src, char *dest, int len)
{
  int idx;

  if (len == 0)len = strlen(src);
  for (idx = 0; idx < len; idx++)
   dest[idx] = toupper(src[idx]);

}

/* -------------------------------------------------------------- */
/* function MassageSimilaritySortWord                             */
/* Helper Function called by ReadSections                         */
/* Strips Carriage Return - Line Feed and other Crapola from      */
/* text file lines...                                             */
/* input(s) : linebuffer                                          */
/* output(s): linebuffer (modified)                               */
/* returns  : none                                                */
/* -------------------------------------------------------------- */
void StripEol(char *ptr)
{
  int idx;
  char c;

  for (idx = 0; ptr[idx] != ASCIIZ; idx++) {
    c = ptr[idx];
    if (c == CARRIAGE_RETURN ||
        c == LINE_FEED ||
        c == CONTROL_Z) {
      ptr[idx] = ASCIIZ;
      break;
    }
  }
}

/* -------------------------------------------------------------- */
/* function ReadSections                                          */
/* Helper Function called by InitializeDictionary                 */
/* Parser Function that does all the work reading metasim.dic     */
/* - if called with readflag set to false pre-reads metasim.dic   */
/* - if called with readflag set to true reads metasim.dic into   */
/* the arrays that have been allocated outside of here.           */
/* input(s) : File Pointer (fp)                                   */
/* output(s): All Values For Dictionary Arrays                    */
/* returns  : total number of elements in all dictionaries        */
/* -------------------------------------------------------------- */
int ReadSections(FILE *fp, int readflag)
{
  unsigned int precnt = 0, sufcnt = 0, subcnt = 0, skipcnt = 0;
  int idx;
  char *ptr;

  rewind(fp);
  CurrentSection = NO_SECTION;

  while (NULL != fgets(szSimBuf, 128, fp)) {
    /* ignore comments */
    if (szSimBuf[0] == '#') continue;
    /* strip carriage returns */
    StripEol(szSimBuf);
    /* ignore blank lines */
    if (szSimBuf[0] == ASCIIZ) continue;

    if (szSimBuf[0] == '[') {
      /* find out what section we have changed to */
      /* and make that the current section */
      CurrentSection = NO_SECTION;
      for (idx = 0; idx < NUM_SECTIONS; idx++) {
        if (!stricmp((char *)&Section[idx][0], szSimBuf)) {
          CurrentSection = idx;
          break;
        }
      }
    }
    else {
      /* build the dictionary if not merely counting */
      if (TRUE == readflag) {
        /* make sure it's upper case */
        StringUpper(szSimBuf, szSimBuf, 0);
        switch(CurrentSection) {
          case PREFIX_SECTION:
            if (numprefixes != 0 && precnt <= numprefixes) {
              szSimBuf[DIC_LEN] = ASCIIZ;
              strcpy((char *)&prefix[precnt].szSimDic[0], szSimBuf);
            }
            break;
          case SUFFIX_SECTION:
            if (numsuffixes != 0 && sufcnt <= numsuffixes) {
              szSimBuf[DIC_LEN] = ASCIIZ;
              strcpy((char *)&suffix[sufcnt].szSimDic[0], szSimBuf);
            }
            break;
          case SUB_SECTION:
            if (numsubs != 0 && subcnt <= numsubs) {
              /* resolve find=replace paired words */
              ptr = (char *)&szSimBuf[0];
              for (idx = 0; szSimBuf[idx] != 0;) {
                if (szSimBuf[idx] == '=') {
                  szSimBuf[idx] = ASCIIZ;
                  ptr = (char *)&szSimBuf[idx+1];
                }
                else
                  idx++;
              }
              ptr[DIC_LEN] = ASCIIZ;
              strcpy((char *)&sub[subcnt].szReplace[0], ptr);
              szSimBuf[DIC_LEN] = ASCIIZ;
              strcpy((char *)&sub[subcnt].szFind[0], szSimBuf);
            }
            break;
          case REMOVE_SECTION:
            if (numskips != 0 && skipcnt <= numskips) {
              szSimBuf[DIC_LEN] = ASCIIZ;
              strcpy((char *)&skip[skipcnt].szSimDic[0], szSimBuf);
            }
            break;
        }
      }
      switch(CurrentSection) {
        case PREFIX_SECTION: precnt++; break;
        case SUFFIX_SECTION: sufcnt++; break;
        case SUB_SECTION:    subcnt++; break;
        case REMOVE_SECTION: skipcnt++;break;
      }
    }
  }
  if (FALSE == readflag) {
    if (numsubs == 0)
      numsubs = subcnt;
    if (numskips == 0)
      numskips = skipcnt;
    if (numsuffixes == 0)
      numsuffixes = sufcnt;
    if (numprefixes == 0)
      numprefixes = precnt;
  }
  return (numsubs + numskips + numsuffixes + numprefixes);

}


/* -------------------------------------------------------------- */
/* function AllocateDictionary                                    */
/* Helper Function called by InitializeDictionary                 */
/* allocates the arrays required for the metasim.dic values       */
/* if memory is insufficient resets initial value so no harm done */
/* input(s) : number of elements required for each array          */
/* output(s): allocated arrays                                    */
/* returns  : total number of elements allocated                  */
/* -------------------------------------------------------------- */
int AllocateDictionary()
{
  if (numsubs) {
    if (NULL ==
      (sub = (WILDSUB *) malloc(sizeof(WILDSUB) * numsubs)))
      numsubs = 0;
  }
  if (numskips) {
    if (NULL ==
      (skip = (WILDDIC *) malloc(sizeof(WILDDIC) * numskips)))
      numskips = 0;
  }
  if (numprefixes) {
    if (NULL ==
      (prefix = (WILDDIC *) malloc(sizeof(WILDDIC) * numprefixes)))
      numprefixes = 0;
  }
  if (numsuffixes) {
    if (NULL ==
      (suffix = (WILDDIC *) malloc(sizeof(WILDDIC) * numsuffixes)))
      numsuffixes = 0;
  }
  return (numsubs + numskips + numsuffixes + numprefixes);
}


/* -------------------------------------------------------------- */
/* function FreeDictionary                                        */
/* Helper Function called by InitializeDictionary and main        */
/* Frees any previous dictionaries and sets initial state for     */
/* reading the metasim.dic file into the arrays                   */
/* input(s) : number of elements in each array if any             */
/* output(s): free and clear initial values for dictionary        */
/* returns  : none                                                */
/* -------------------------------------------------------------- */
void FreeDictionary()
{
  if (numsubs)free(sub);
  if (numskips)free(skip);
  if (numprefixes)free(prefix);
  if (numsuffixes)free(suffix);

  numsubs = 0;
  numskips = 0;
  numprefixes = 0;
  numsuffixes = 0;

  sub     = (WILDSUB *)NULL;
  skip    = (WILDDIC *)NULL;
  prefix  = (WILDDIC *)NULL;
  suffix  = (WILDDIC *)NULL;

}


/* -------------------------------------------------------------- */
/* function InitializeDictionary                                  */
/* Helper Function called by main()                               */
/* reads and parses the dictionary file and creates the arrays    */
/* required to perform language-specific substitution etc.        */
/* input(s) : dictionary filename                                 */
/* output(s): dictionary arrays                                   */
/* returns  : none                                                */
/* -------------------------------------------------------------- */
void InitializeDictionary(char *szDictName)
{
  FILE *fp;

  FreeDictionary();

  if (NULL != (fp = fopen(szDictName, "r"))) {
    /* count subs */
    /* count skips */
    /* count prefixes */
    /* count suffixes */
    if (ReadSections(fp, FALSE)) {
      /* allocate subs */
      /* allocate skips */
      /* allocate prefixes */
      /* allocate suffixes */
      if (AllocateDictionary()) {
        /* build sub array */
        /* build skip array */
        /* build prefix array */
        /* build suffix array */
        ReadSections(fp, TRUE);
      }
    }
    /* close the file */
    fclose(fp);
  }
}

/* start of filter program user interface (main() program) */
/* contructed to provide as many input and output methods */
/* as possible */

/* function main() - entry point for input */

int main(int argc, char **argv)
{
  char c,
       *ptr = NULL;
  FILE *fp1 = NULL,
       *fp2 = NULL,
       *fp3 = NULL;

  long lpos = 0L;
  int idx = 0, shift = 0, len;
  unsigned int wildcount = 0;
  int iInputMethod = INPUT_STREAM;

  /* step 1. setup for input method */
  /* default input method is stream */
  /* this can be either from a redirected file */
  /* or from console (stdin) */
  /* other possible input methods are */
  /* from file (named on command line) */
  /* or from phrases supplied as command line args */

  /* open the dictionary if it exists and */
  /* create the arrays of words to process */
  InitializeDictionary(szDicFile);

  /* create the sort array for reordering of sentence phrases */
  /* allocate a buffer for 100 words x 128 characters long */

  if((wildword=malloc(sizeof(WILDWORD)*WORD_MAX))==NULL) {
    fprintf(stderr, "Memory Allocation Error!");
    return (1);
  }

  /* parse command line switches... if they exist... */
  /* command line switches must always be first arg on commandline */
  /* and must always be preceded by a switch-char */
  if (argc > 1) {
    if (argv[1][0] == '-' || argv[1][0] == '/') {
      shift = 1;
      for (idx = 1; argv[1][idx] != ASCIIZ; idx++) {
        c = toupper(argv[1][idx]);
        switch (c) {
          case 'S': bSort = TRUE; break;
          case 'R': bReverse = TRUE; break;
          case 'V': bVerbatim = TRUE; break;
          case 'D': bDebug = TRUE; break;
        }
      }
    }
  }

  /* allocate memory if internal sort is true */
  /* we will sort the keys in memory and use a temp sort file */
  /* to store the real string... */
  if (TRUE == bSort) {
    wildphrase = (WILDPHRASE *)malloc(KEY_MAX * sizeof(WILDPHRASE));
    if (NULL == wildphrase) {
      bSort = FALSE;
    }
  }

  /* open internal sort file if internal sort is true */
  if (TRUE == bSort) {
    fp3 = fopen(szTempSortFile, "wb+r");
    if (fp3 == NULL) {
      bSort = FALSE;
      free(wildphrase);
    }
  }

  if((argc-shift) < 2)
  {
    iInputMethod = INPUT_STREAM;
    /* if input is from stream, prompt from stderr */
    /* so we don't "pollute" stdout */
    fprintf(stderr, pszCredit);
    fprintf(stderr, pszPurpose);
    fprintf(stderr, pszUsage);
    fprintf(stderr, pszPrompt);
  }
  else {
    if(NULL != (fp1 = fopen(argv[1+shift], "r")))
    {
      iInputMethod = INPUT_FILE;
      if((argc-shift) > 2)
        fp2 = fopen(argv[2+shift], "w");
    }
    else
      iInputMethod = INPUT_ARGV;
  }
  
  /* step 2. setup is complete so proceed with metaphone */
  /* with similarity sort key */
  /* and direct to stdout or to file... */

  for(;;) 
  {
    /* step 2.1 */
    /* check for input method and based on how we getting input */
    /* read the next phrase to be converted into the input buffer */
    /* and create the similarity sort key copy */
    /* before calling metaphone */
    
    switch(iInputMethod) 
    {
      case INPUT_ARGV:
        idx++;
        if(NULL != (ptr = argv[idx+shift]))
          strcpy(szName, ptr);
        break;
      case INPUT_FILE:
        if(NULL != (ptr = fgets(szName, 128, fp1)))
        {
          for(idx = 0; szName[idx] != ASCIIZ;)
          {
            if(szName[idx] == CARRIAGE_RETURN ||
               szName[idx] == LINE_FEED)
              szName[idx] = ASCIIZ;
            else
              idx++;
          }
        }
        break;
      case INPUT_STREAM:
      default:
        ptr = gets(szName);
        break;
    }
    
    /* step 2.2 */
    /* break if end-of-input otherwise print output */
    
    if (NULL == ptr ||
        szName[0] == CARRIAGE_RETURN ||
        szName[0] == LINE_FEED ||
        szName[0] == CONTROL_Z ||
        szName[0] == ASCIIZ) {

      if (NULL == ptr || szName[0] == CONTROL_Z)
        break;

      continue;
    }

    /* create the new Similarity Sort Key copy */
    CreateSimilaritySortKey(szName, szSimBuf);

    /* work on the copy */
    /* if we are not using metaphone to sort */
    /* use the squished version of the string for our sort */
    if (bVerbatim == TRUE) {
      sprintf(szMetaph, "%-*s", META_LEN, szSimBuf);
      szMetaph[META_LEN] = ASCIIZ;
    }
    else
      phonetic(szSimBuf, szMetaph, META_LEN);

    if (TRUE == bSort) {
      if (wildcount < KEY_MAX) {
        /* write the real string to file */
        len = strlen(szName);
        fwrite(&len, LEN_LEN, 1, fp3);
        fwrite(szName, len + 1, 1, fp3);
        /* add the key and position to the sort array */
        sprintf((char *)&wildphrase[wildcount].szSortKey[0],
                "%-*s", META_LEN, szMetaph);
        wildphrase[wildcount].lpos = lpos;
        lpos = (lpos + LEN_LEN + len + 1);
        wildcount++;
        if (wildcount == KEY_MAX)
          fprintf(stderr, "Maximum Internal Sort Keys %d reached...",
            KEY_MAX);
      }
    }
    else {

      if(fp2 == NULL)
        printf("%-*s=%s\n", META_LEN, szMetaph, szName);
      else
        fprintf(fp2, "%-*s=%s\n", META_LEN, szMetaph, szName);
    }
  }

  /* if using internal sort then use the sorted array */
  /* to read the temporary sort file */

  if (NULL != fp3 && wildcount > 0) {
    /* rewind temp file before changing to read mode */
    rewind(fp3);
    if (wildcount > 1) {
      sortsize = META_LEN;
      qsort(wildphrase, wildcount, sizeof(WILDPHRASE), compare);
    }

    szMetaph[META_LEN] = ASCIIZ; /* for debug mode */

    for (idx = 0; idx < wildcount; idx++) {
      fseek(fp3, wildphrase[idx].lpos, SEEK_SET);
      fread(&len, LEN_LEN, 1, fp3);
      fread(szName, len + 1, 1, fp3);
      if (bDebug == FALSE) {
        if(fp2 == NULL)
          printf("%s\n", szName);
        else
          fprintf(fp2, "%s\n", szName);
      }
      else  {
        memcpy(szMetaph, (char *)&wildphrase[idx].szSortKey[0],
               META_LEN);
        if(fp2 == NULL)
          printf("%-*s=%s\n", META_LEN, szMetaph, szName);
        else
         fprintf(fp2, "%-*s=%s\n", META_LEN, szMetaph, szName);
      }
    }
  }

  /* all done so clean-up */

  if(NULL != fp1) fclose(fp1);  /* close any open files */
  if(NULL != fp2) fclose(fp2);
  if(NULL != fp3) {
    /* close and remove temporary sort file */
    fclose(fp3);
    remove(szTempSortFile);
  }

  FreeDictionary(); /* free dictionary arrays */

  free(wildword); /* free the word sort array */

  if (TRUE == bSort)
    free(wildphrase); /* free the phrase sort array */

  return(0);
}
